@@ -1,4 +1,4 @@
-// Copyright (C) 2010-2022 Joel Rosdahl and other contributors
+// Copyright (C) 2010-2023 Joel Rosdahl and other contributors
 //
 // See doc/AUTHORS.adoc for a complete list of contributors.
 //
@@ -45,8 +45,6 @@
 // The option only affects compilation; not passed to the preprocessor.
 #define AFFECTS_COMP (1 << 6)
 
-#define ARRAY_SIZE(array) (sizeof(array) / sizeof((array)[0]))
-
 struct CompOpt
 {
   const char* name;
@@ -98,8 +96,6 @@ const CompOpt compopts[] = {
   {"-Xpreprocessor", AFFECTS_CPP | TOO_HARD_DIRECT | TAKES_ARG},
   {"-Yc", TAKES_ARG | TOO_HARD},                                    // msvc
   {"-Yu", AFFECTS_CPP | TAKES_ARG | TAKES_CONCAT_ARG | TAKES_PATH}, // msvc
-  {"-ZI", TOO_HARD},                                                // msvc
-  {"-Zi", TOO_HARD},                                                // msvc
   {"-all_load", AFFECTS_COMP},
   {"-analyze", TOO_HARD}, // Clang
   {"-arch", TAKES_ARG},
@@ -110,6 +106,8 @@ const CompOpt compopts[] = {
   {"-ccbin", AFFECTS_CPP | TAKES_ARG}, // nvcc
   {"-emit-pch", AFFECTS_COMP},         // Clang
   {"-emit-pth", AFFECTS_COMP},         // Clang
+  {"-external:I",
+   AFFECTS_CPP | TAKES_ARG | TAKES_CONCAT_ARG | TAKES_PATH}, // msvc
   {"-fno-working-directory", AFFECTS_CPP},
   {"-fplugin=libcc1plugin", TOO_HARD}, // interaction with GDB
   {"-frepo", TOO_HARD},
@@ -177,11 +175,8 @@ find(const std::string& option)
 {
   CompOpt key;
   key.name = option.c_str();
-  void* result = bsearch(&key,
-                         compopts,
-                         ARRAY_SIZE(compopts),
-                         sizeof(compopts[0]),
-                         compare_compopts);
+  void* result = bsearch(
+    &key, compopts, std::size(compopts), sizeof(compopts[0]), compare_compopts);
   return static_cast<CompOpt*>(result);
 }
 
@@ -192,7 +187,7 @@ find_prefix(const std::string& option)
   key.name = option.c_str();
   void* result = bsearch(&key,
                          compopts,
-                         ARRAY_SIZE(compopts),
+                         std::size(compopts),
                          sizeof(compopts[0]),
                          compare_prefix_compopts);
   return static_cast<CompOpt*>(result);
@@ -205,7 +200,7 @@ bool compopt_verify_sortedness_and_flags();
 bool
 compopt_verify_sortedness_and_flags()
 {
-  for (size_t i = 0; i < ARRAY_SIZE(compopts); i++) {
+  for (size_t i = 0; i < std::size(compopts); ++i) {
     if (compopts[i].type & TOO_HARD && compopts[i].type & TAKES_CONCAT_ARG) {
       PRINT(stderr,
             "type (TOO_HARD | TAKES_CONCAT_ARG) not allowed, used by {}\n",
