@@ -1,4 +1,4 @@
-// Copyright (C) 2021-2022 Joel Rosdahl and other contributors
+// Copyright (C) 2021-2023 Joel Rosdahl and other contributors
 //
 // See doc/AUTHORS.adoc for a complete list of contributors.
 //
@@ -76,12 +76,12 @@ const StatisticsField k_statistics_fields[] = {
   FIELD(cleanups_performed, nullptr),
   FIELD(compile_failed, "Compilation failed", FLAG_UNCACHEABLE),
   FIELD(compiler_check_failed, "Compiler check failed", FLAG_ERROR),
+  FIELD(compiler_produced_no_output,
+        "Compiler output file missing",
+        FLAG_UNCACHEABLE),
   FIELD(compiler_produced_empty_output,
         "Compiler produced empty output",
         FLAG_UNCACHEABLE),
-  FIELD(compiler_produced_no_output,
-        "Compiler produced no output",
-        FLAG_UNCACHEABLE),
   FIELD(compiler_produced_stdout, "Compiler produced stdout", FLAG_UNCACHEABLE),
   FIELD(could_not_find_compiler, "Could not find compiler", FLAG_ERROR),
   FIELD(could_not_use_modules, "Could not use modules", FLAG_UNCACHEABLE),
@@ -90,6 +90,7 @@ const StatisticsField k_statistics_fields[] = {
         FLAG_UNCACHEABLE),
   FIELD(direct_cache_hit, nullptr),
   FIELD(direct_cache_miss, nullptr),
+  FIELD(disabled, "Ccache disabled", FLAG_UNCACHEABLE),
   FIELD(error_hashing_extra_file, "Error hashing extra file", FLAG_ERROR),
   FIELD(files_in_cache, nullptr, FLAG_NOZERO),
   FIELD(internal_error, "Internal error", FLAG_ERROR),
@@ -127,10 +128,14 @@ const StatisticsField k_statistics_fields[] = {
   FIELD(unsupported_source_language,
         "Unsupported source language",
         FLAG_UNCACHEABLE),
+
+  // subdir_files_base and subdir_size_kibibyte_base are intentionally omitted
+  // since they are not interesting to show.
 };
 
 static_assert(sizeof(k_statistics_fields) / sizeof(k_statistics_fields[0])
-              == static_cast<size_t>(Statistic::END) - 1);
+              == static_cast<size_t>(Statistic::END)
+                   - (/*none*/ 1 + /*subdir files*/ 16 + /*subdir size*/ 16));
 
 static std::string
 format_timestamp(const util::TimePoint& value)
@@ -152,10 +157,13 @@ percent(const uint64_t nominator, const uint64_t denominator)
 {
   if (denominator == 0) {
     return "";
-  } else if (nominator >= denominator) {
-    return FMT("({:5.1f}%)", (100.0 * nominator) / denominator);
+  }
+
+  std::string result = FMT("({:5.2f}%)", (100.0 * nominator) / denominator);
+  if (result.length() <= 8) {
+    return result;
   } else {
-    return FMT("({:5.2f}%)", (100.0 * nominator) / denominator);
+    return FMT("({:5.1f}%)", (100.0 * nominator) / denominator);
   }
 }
 
@@ -298,7 +306,13 @@ Statistics::format_human_readable(const Config& config,
     add_ratio_row(table, "  Preprocessed:", p_hits, p_hits + p_misses);
   }
 
-  const uint64_t g = 1'000'000'000;
+  const char* size_unit =
+    config.size_unit_prefix_type() == util::SizeUnitPrefixType::binary ? "GiB"
+                                                                       : "GB";
+  const uint64_t size_divider =
+    config.size_unit_prefix_type() == util::SizeUnitPrefixType::binary
+      ? 1024 * 1024 * 1024
+      : 1000 * 1000 * 1000;
   const uint64_t local_hits = S(local_storage_hit);
   const uint64_t local_misses = S(local_storage_miss);
   const uint64_t local_reads =
@@ -314,21 +328,24 @@ Statistics::format_human_readable(const Config& config,
   const uint64_t remote_errors = S(remote_storage_error);
   const uint64_t remote_timeouts = S(remote_storage_timeout);
 
-  table.add_heading("Local storage:");
+  if (!from_log || verbosity > 0 || (local_hits + local_misses) > 0) {
+    table.add_heading("Local storage:");
+  }
   if (!from_log) {
     std::vector<C> size_cells{
-      "  Cache size (GB):",
-      C(FMT("{:.2f}", static_cast<double>(local_size) / g)).right_align()};
+      FMT("  Cache size ({}):", size_unit),
+      C(FMT("{:.1f}", static_cast<double>(local_size) / size_divider))
+        .right_align()};
     if (config.max_size() != 0) {
       size_cells.emplace_back("/");
       size_cells.emplace_back(
-        C(FMT("{:.2f}", static_cast<double>(config.max_size()) / g))
+        C(FMT("{:.1f}", static_cast<double>(config.max_size()) / size_divider))
           .right_align());
       size_cells.emplace_back(percent(local_size, config.max_size()));
     }
     table.add_row(size_cells);
 
-    if (verbosity > 0) {
+    if (verbosity > 0 || config.max_files() > 0) {
       std::vector<C> files_cells{"  Files:", S(files_in_cache)};
       if (config.max_files() > 0) {
         files_cells.emplace_back("/");
@@ -342,7 +359,7 @@ Statistics::format_human_readable(const Config& config,
       table.add_row({"  Cleanups:", cleanups});
     }
   }
-  if (verbosity > 0 || (remote_hits + remote_misses) > 0) {
+  if (verbosity > 0 || (local_hits + local_misses) > 0) {
     add_ratio_row(table, "  Hits:", local_hits, local_hits + local_misses);
     add_ratio_row(table, "  Misses:", local_misses, local_hits + local_misses);
   }
