@@ -1,4 +1,4 @@
-// Copyright (C) 2021-2022 Joel Rosdahl and other contributors
+// Copyright (C) 2021-2023 Joel Rosdahl and other contributors
 //
 // See doc/AUTHORS.adoc for a complete list of contributors.
 //
@@ -255,6 +255,9 @@ read_file_part(const std::string& path, size_t pos, size_t count)
 template nonstd::expected<util::Bytes, std::string>
 read_file_part(const std::string& path, size_t pos, size_t count);
 
+template nonstd::expected<std::string, std::string>
+read_file_part(const std::string& path, size_t pos, size_t count);
+
 template nonstd::expected<std::vector<uint8_t>, std::string>
 read_file_part(const std::string& path, size_t pos, size_t count);
 
@@ -296,7 +299,7 @@ nonstd::expected<void, std::string>
 write_fd(int fd, const void* data, size_t size)
 {
   int64_t written = 0;
-  do {
+  while (static_cast<size_t>(written) < size) {
     const auto count =
       write(fd, static_cast<const uint8_t*>(data) + written, size - written);
     if (count == -1) {
@@ -306,7 +309,7 @@ write_fd(int fd, const void* data, size_t size)
     } else {
       written += count;
     }
-  } while (static_cast<size_t>(written) < size);
+  }
   return {};
 }
 
